"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = _default;

var _dlv = _interopRequireDefault(require("dlv"));

var _postcss = _interopRequireDefault(require("postcss"));

var _node = _interopRequireDefault(require("postcss/lib/node"));

var _escapeClassName = _interopRequireDefault(require("./escapeClassName"));

var _generateVariantFunction = _interopRequireDefault(require("./generateVariantFunction"));

var _parseObjectStyles = _interopRequireDefault(require("./parseObjectStyles"));

var _prefixSelector = _interopRequireDefault(require("./prefixSelector"));

var _wrapWithVariants = _interopRequireDefault(require("./wrapWithVariants"));

var _cloneNodes = _interopRequireDefault(require("./cloneNodes"));

var _transformThemeValue = _interopRequireDefault(require("./transformThemeValue"));

var _nameClass = _interopRequireDefault(require("./nameClass"));

var _isKeyframeRule = _interopRequireDefault(require("./isKeyframeRule"));

var _toPath = require("./toPath");

var _defaults = require("./defaults");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function parseStyles(styles) {
  if (!Array.isArray(styles)) {
    return parseStyles([styles]);
  }

  return styles.flatMap(style => style instanceof _node.default ? style : (0, _parseObjectStyles.default)(style));
}

function wrapWithLayer(rules, layer) {
  return _postcss.default.atRule({
    name: 'layer',
    params: layer
  }).append((0, _cloneNodes.default)(Array.isArray(rules) ? rules : [rules]));
}

function _default(plugins, config) {
  const pluginBaseStyles = [];
  const pluginComponents = [];
  const pluginUtilities = [];
  const pluginVariantGenerators = {};

  const applyConfiguredPrefix = selector => {
    return (0, _prefixSelector.default)(config.prefix, selector);
  };

  function addUtilities(utilities, options) {
    const defaultOptions = {
      variants: [],
      respectPrefix: true,
      respectImportant: true
    };
    options = Array.isArray(options) ? Object.assign({}, defaultOptions, {
      variants: options
    }) : (0, _defaults.defaults)(options, defaultOptions);

    const styles = _postcss.default.root({
      nodes: parseStyles(utilities)
    });

    styles.walkRules(rule => {
      if (options.respectPrefix && !(0, _isKeyframeRule.default)(rule)) {
        rule.selector = applyConfiguredPrefix(rule.selector);
      }

      if (options.respectImportant && config.important) {
        rule.__tailwind = { ...rule.__tailwind,
          important: config.important
        };
      }
    });
    pluginUtilities.push(wrapWithLayer((0, _wrapWithVariants.default)(styles.nodes, options.variants), 'utilities'));
  }

  const getConfigValue = (path, defaultValue) => path ? (0, _dlv.default)(config, path, defaultValue) : config;

  plugins.forEach(plugin => {
    var _plugin$handler, _plugin;

    if (plugin.__isOptionsFunction) {
      plugin = plugin();
    }

    const handler = typeof plugin === 'function' ? plugin : (_plugin$handler = (_plugin = plugin) === null || _plugin === void 0 ? void 0 : _plugin.handler) !== null && _plugin$handler !== void 0 ? _plugin$handler : () => {};
    handler({
      postcss: _postcss.default,
      config: getConfigValue,
      theme: (path, defaultValue) => {
        let [pathRoot, ...subPaths] = (0, _toPath.toPath)(path);
        let value = getConfigValue(['theme', pathRoot, ...subPaths], defaultValue);
        return (0, _transformThemeValue.default)(pathRoot)(value);
      },
      corePlugins: path => {
        if (Array.isArray(config.corePlugins)) {
          return config.corePlugins.includes(path);
        }

        return getConfigValue(`corePlugins.${path}`, true);
      },
      variants: (path, defaultValue) => {
        if (Array.isArray(config.variants)) {
          return config.variants;
        }

        return getConfigValue(`variants.${path}`, defaultValue);
      },
      e: _escapeClassName.default,
      prefix: applyConfiguredPrefix,
      addUtilities,
      matchUtilities: (matches, {
        values,
        variants,
        respectPrefix,
        respectImportant
      }) => {
        let modifierValues = Object.entries(values || {});
        let result = Object.entries(matches).flatMap(([name, utilityFunction]) => {
          return modifierValues.map(([modifier, value]) => {
            let declarations = utilityFunction(value, {
              includeRules(rules, options) {
                addUtilities(rules, options);
              }

            });

            if (!declarations) {
              return null;
            }

            return {
              [(0, _nameClass.default)(name, modifier)]: declarations
            };
          }).filter(Boolean);
        });
        addUtilities(result, {
          variants,
          respectPrefix,
          respectImportant
        });
      },
      addComponents: (components, options) => {
        const defaultOptions = {
          variants: [],
          respectPrefix: true
        };
        options = Array.isArray(options) ? Object.assign({}, defaultOptions, {
          variants: options
        }) : (0, _defaults.defaults)(options, defaultOptions);

        const styles = _postcss.default.root({
          nodes: parseStyles(components)
        });

        styles.walkRules(rule => {
          if (options.respectPrefix && !(0, _isKeyframeRule.default)(rule)) {
            rule.selector = applyConfiguredPrefix(rule.selector);
          }
        });
        pluginComponents.push(wrapWithLayer((0, _wrapWithVariants.default)(styles.nodes, options.variants), 'components'));
      },
      addBase: baseStyles => {
        pluginBaseStyles.push(wrapWithLayer(parseStyles(baseStyles), 'base'));
      },
      addVariant: (name, generator, options = {}) => {
        pluginVariantGenerators[name] = (0, _generateVariantFunction.default)(generator, options);
      }
    });
  });
  return {
    base: pluginBaseStyles,
    components: pluginComponents,
    utilities: pluginUtilities,
    variantGenerators: pluginVariantGenerators
  };
}